#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    CLI support for restretto
    ~~~~~~~~~~~~~~~~~~~~~~~~~
"""


import sys
from argparse import ArgumentParser, ArgumentTypeError
from . import load
from .errors import ExpectError
from clint.textui import colored, puts


def options(encoded):
    """Returns dict parsed from encoded string in form key1=val1,key2=val2"""
    opts = {}
    try:
        parts = (p.strip().split("=") for p in encoded.split(",") if p.strip())
        for key, val in parts:
            if not key.strip():
                raise ValueError
            opts[key.strip()] = val.strip()
    except Exception:
        raise ArgumentTypeError(encoded)
    return opts


parser = ArgumentParser(description="REST resources/endpoints testing tool")
parser.add_argument("path", help="path to look for tests (file or directory)")
#parser.add_argument("--xunit", dest="xunit_dir", default=None,
#                    help="output xunit reports to this dir")
parser.add_argument("--print-passed", action="store_true", help="Print passed tests")
parser.add_argument("--print-response", action="store_true", help="Print responses")
parser.add_argument(
    "--vars", action="store", type=options,
    help="Context variables as var1=val1,var2=val2")
parser.add_argument(
    "--debug-errors", action="store_true",
    help="Open ipdb (should be isntalled) debugger on errors"
)


def main(args=sys.argv[1:]):
    arguments = parser.parse_args(args)

    sessions = load(arguments.path)
    if not sessions:
        print("No test sessions found, exiting")
        sys.exit(1)

    passed = failed = errors = 0
    for test_session in sessions:
        hdr = "Test session: {}".format(test_session.title)
        print(hdr)
        print('-' * len(hdr))
        for resource in test_session.resources:
            try:
                test_session.test(resource, context=arguments.vars)
                if arguments.print_passed:
                    # TODO: print response status instead
                    print("{} {}: Ok".format(colored.green("[PASS]"), resource.title))
                if arguments.print_response:
                    print(resource.response.text)
                passed += 1
            except ExpectError as failure:
                print("{} {}: {}".format(colored.red("[FAIL]"), resource.title, failure))
                if arguments.print_response:
                    print(resource.response.text)
                failed += 1
            except Exception as error:
                print("{} {}: {}".format(colored.yellow("[ERROR]"), resource.title, error))
                # TODO: remove it
                if arguments.debug_errors:
                    import ipdb
                    ipdb.set_trace()
                errors += 1
        print("")
    totals = "Total: {} / Passed: {} / Errors: {} / Failed: {}".format(
        passed+failed+errors, colored.green(passed), colored.yellow(errors), colored.red(failed)
    )
    print("-" * len(totals))
    print(totals)
    print("")
    return 1 if (failed or errors) else 0




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import requests
from urllib.request import urljoin
from . import templating
from . import assertions

logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)


class Result(object):

    def __init__(self, title=None):
        self.title = title
        self.failures = []
        self.succeed = []

    @property
    def ok(self):
        return len(self.failures) == 0


class Runner(object):

    def __init__(self, spec, context={}):
        # define context
        self.context = spec.get('vars', {})
        self.context.update(context)
        # apply context
        self.spec = templating.apply_session_context(spec, self.context)
        # setup session
        self.session = requests.Session()
        self.session.headers.update(self.spec.get('headers', {}))
        self.baseUri = spec.get('baseUri', '')
        # prepare result
        self.result = Result(self.spec.get('title', ''))

    @property
    def actions(self):
        return self.spec['actions']

    def execute(self, action):
        # apply context for request
        spec = templating.apply_action_context(action, self.context)
        # create assertions
        assertion = assertions.Assert(spec.pop('assert') if 'assert' in spec else [])
        # get response
        response = self.request(**spec)
        # test assertion
        try:
            assertion.test(response)
            self.result.succeed.append((response, None))
            return self.result.succeed[-1]
        except AssertionError as error:
            self.result.failures.append((response, error))
            return self.result.failures[-1]
        # TODO: store vars in context

    def request(self, **spec):
        spec['url'] = urljoin(self.baseUri, spec['url'])
        request = requests.Request(**spec)
        prepared = request.prepare()
        return self.session.send(prepared)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import requests
from urllib.request import urljoin


class Session(object):
    """REST session"""

    def __init__(self, headers={}, baseUri='', **kwargs):
        self.session = requests.Session()
        # set common headers
        self.session.headers.update(headers)
        self.baseUri = baseUri

    def action(self, url, method='GET', headers=None, body=None, json=None):
        # create Request
        request = requests.Request(method, urljoin(self.baseUri, url), headers=headers)
        if json:
            request.json = json
        elif body:
            request.data = body
        prepared = request.prepare()
        return self.session.send(prepared)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import yaml
from jinja2 import Template


def apply_context(src, context={}):
    """Apply context to dict"""
    return yaml.load(Template(yaml.dump(src)).render(context))


def json_path(path, data):
    """Extract property by path"""
    fragments = path.split(".")
    src = data
    for p in fragments:
        src = src.get(p, {})
    return src



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    restretto - REST resource endpoints testing toolkit
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from . import assertions
from . import errors
from . import utils
from .rest import Resource, Session
from .loader import load



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from fnmatch import fnmatch
from .utils import json_path
from .errors import ExpectError


class ResponseTest(object):

    message = "Bad response ({0} {1})"

    def expect(self, statement, message=''):
        if not statement:
            raise ExpectError(message)

    def test(self, response):
        self.expect(response, self.message.format(response.status_code, response.reason))

    def assert_is(self, item, value):
        # coerce int and str types
        types = set((type(item), type(value)))
        if types == set((str, int)):
            cond = str(item) == str(value)
        else:
            cond = item == value
        self.expect(cond, "{} != {}".format(item, value))

    def assert_is_not(self, item, value):
        # coerce int and str types
        types = set((type(item), type(value)))
        if types == set((str, int)):
            cond = str(item) != str(value)
        else:
            cond = item != value
        self.expect(cond, "{} == {}".format(item, value))

    def assert_contains(self, item, value):
        self.expect(value in item, "{} not found in {}".format(value, item))

    def assert_statements(self, statements, item):
        # assert all conditions are satisfied
        for (cond, value) in statements.items():
            assert_fn = 'assert_{}'.format(cond)
            getattr(self, assert_fn)(item, value)


class StatusCodeTest(ResponseTest):

    message = "Unexpected status ({0} instead of {1})"

    def __init__(self, status):
        # convert int code to str
        self.expected = str(status) if isinstance(status, int) else status

    def test(self, response):
        if isinstance(self.expected, str):
            self.expect(
                fnmatch(str(response.status_code), self.expected.replace('x', '?')),
                self.message.format(response.status_code, self.expected)
            )
        else:
            # treat as list of possible statuses
            self.expect(
                str(response.status_code) in self.expected,
                self.message.format(response.status_code, self.expected)
            )


class ResponsePropertyTest(ResponseTest):

    def __init__(self, name, statements={}):
        self.name = name
        self.statements = statements


class HeaderTest(ResponsePropertyTest):

    def test(self, response):
        header = response.headers.get(self.name, None)
        self.expect(header, "Header not found: {}".format(self.name))
        self.assert_statements(self.statements, header)


class BodyTest(ResponsePropertyTest):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # pop property definition, if available
        self.prop = self.statements.pop('property', None)

    def test(self, response):
        data = None
        if self.name == 'text':
            data = response.text
        elif self.name == 'json' and not self.prop:
            data = response.json()
        elif self.name == 'json' and self.prop:
            # get required property value
            data = json_path(self.prop, {'json': response.json()})
        self.expect(data, "Content not found or empty: {}".format(self.name))
        self.assert_statements(self.statements, data)


class Assert(object):

    def __init__(self, statements=[]):
        self.statements = []
        self._has_status_test = False
        if not statements:
            # assume default simple check
            self.statements = [ResponseTest()]
        else:
            for spec in statements:
                self.statements.append(self.statement(spec))
            if not self._has_status_test:
                # add default test for status_code to be ok
                self.statements.insert(0, ResponseTest())

    def test(self, response):
        for stmt in self.statements:
            stmt.test(response)
        return True

    def statement(self, spec):
        """Statement factory"""
        spec = dict(spec)
        if 'status' in spec:
            self._has_status_test = True
            return StatusCodeTest(spec['status'])
        if 'header' in spec:
            return HeaderTest(spec.pop('header'), spec)
        if 'body' in spec:
            return BodyTest(spec.pop('body'), spec)



# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

setup(
    name="restretto",
    version="0.32",
    description="restretto is REST API testing tool",
    #long_description=description,
    author="Arthur Orlov",
    author_email="feature.creature@gmail.com",
    url="http://github.com/wirewit/restretto",
    license="",
    packages=find_packages(),
    entry_points={"console_scripts": ["restretto = restretto.cli:main"]},
    install_requires=["requests>=2.7.0", "pyaml>=3.11", "jinja2>=2.8", "clint>=0.5"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "License :: OSI Approved :: Apache Software License"
    ]
)



# -*- coding: utf-8 -*-
"""
    Core classes for restretto
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import requests
from urllib.request import urljoin

from .utils import json_path
from . import assertions
from .errors import ParseError
from .utils import apply_context


HTTP_METHODS = frozenset(('get', 'options', 'head', 'post', 'put', 'patch', 'delete'))


class Resource(object):
    """Single HTTP resource"""

    @staticmethod
    def parse_from_dict(spec):
        """Expand from shortened forms to standart form"""
        # guess method
        request = {
            'url': spec.get('url'),
            'method': spec.get('method'),
            'headers': spec.get('headers'),
            'params': spec.get('params'),
            'data': spec.get('data'),
            'json': spec.get('json')
        }
        if request['url'] and not request['method']:
            # if url is given, assume methid is get
            request['method'] = 'get'
        elif not request['method']:
            # search method defintion
            methods = set(spec.keys()).intersection(HTTP_METHODS)
            if len(methods) != 1:
                raise ParseError('Multiple methods given for action')
            http_verb = methods.pop()
            # get url
            request['url'] = spec[http_verb]
            request['method'] = http_verb
        # validate fields
        if not request['url'] or not request['method']:
            raise ParseError('Url or method for action not specified')
        request['method'] = request['method'].lower()
        if request['method'] not in HTTP_METHODS:
            raise ParseError('Unknown http method verb: {}'.format(spec['method']))
        # clean empty fields
        return {k: v for k, v in request.items() if v is not None}

    def __init__(self, spec):
        """Create resource from specification"""
        if isinstance(spec, str):
            self.spec = {'url': spec}
        else:
            self.spec = spec

        # get context var bindings
        self.vars = self.spec.get('vars', {})

        # get asserions
        if 'expect' in spec and 'assert' in spec:
            # only one form of assertions should be used at a time
            raise ParseError("Only expect or assert keyword can be used")
        self.asserts = self.spec.get('assert', self.spec.get('expect'))

        self.request = self.parse_from_dict(self.spec)
        # response and errors are not known
        self.response = None
        self.error = None

    @property
    def title(self):
        return self.spec.get('title') or self.spec.get('name') \
            or '{method} {url}'.format(**self.request)

    def test(self, baseUri='', context={}, session=None):
        """Make request, perform assertion testing"""
        self.request['url'] = urljoin(baseUri, self.request['url'].lstrip('/'))
        # apply template to request and assertions
        self.request = apply_context(self.request, context)
        self.asserts = apply_context(self.asserts, context)
        # create assertions
        assertion = assertions.Assert(self.asserts)
        # get response
        http = session or requests.Session()
        self.response = http.request(**self.request)
        # test assertion, will raise an excep
        try:
            assertion.test(self.response)
        except Exception as error:
            # save error
            self.error = error
            # reraise
            raise
        # save context vars
        if self.vars:
            try:
                data = {
                    'json': self.response.json(),
                    'headers': self.response.headers
                }
                for name, path in self.vars.items():
                    self.vars[name] = json_path(path, data)
            except ValueError:
                # no json, it's can be ok
                pass
        return self


class Session(object):
    """REST session"""

    def __init__(self, spec, context={}):
        self.spec = spec
        self.context = spec.get('vars', {}).copy()
        self.context.update(context)
        self.baseUri = apply_context(spec.get('baseUri', ''), self.context)
        self.http = requests.Session()
        headers = self.spec.get('headers') or {}
        self.headers = apply_context(headers, self.context)
        self.http.headers.update(self.headers)
        # create resources
        self.resources = []
        self._parse_resources()

    def _parse_resources(self):
        """Get resources from loaded session spec"""
        entries = self.spec.get('resources') or []
        for item in entries:
            self.resources.append(Resource(item))

    def __bool__(self):
        return bool(self.resources)

    @property
    def filename(self):
        return self.spec.get('filename')

    @property
    def title(self):
        return self.spec.get('title', '') or self.spec.get('name', '') or self.spec.get('session', '')

    def test(self, resource=None, context=None):
        context = context or {}
        context.update(self.context)
        executed = resource.test(self.baseUri, context, self.http)
        self.context.update(executed.vars)
        return executed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
    restretto YML loader
    ~~~~~~~~~~~~~~~~~~~~
"""

import yaml
import os
from .rest import Session


SUPPORTED_EXTENSIONS = (".yml", ".yaml")


def load(path):
    data = []
    files = []
    if os.path.isdir(path):
        # load only files with supported extension skipping hiddens like '.yml'
        for entry in os.listdir(path):
            (name, ext) = os.path.splitext(entry)
            if name and ext in SUPPORTED_EXTENSIONS:
                files.append(os.path.join(path, entry))
    else:
        files.append(path)
    for entry in files:
        with open(entry) as source:
            parsed = yaml.load(source)
        # silently skip empty files
        if not parsed:
            continue
        # add filename to spec
        parsed['filename'] = entry
        data.append(Session(parsed))
    # filter out empty elements (loaded from empty files)
    return [item for item in data if item]
